// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibus;

/// Asynchronous processes helper
///
/// This class helps to await multiple asynchronous processes.
class AsyncProcesses<T> {
  /// Evaluated responses
  List<T> _completed = new List<T>();

  /// Future responses
  List<Future<T>> _futures;

  /// Total number of responses
  int get count => _completed.length + _futures.length;

  /// Completer that completes when all future responses are received.
  Completer<List<T>> _completer = new Completer<List<T>>();

  /// Await the evaluation of all processes
  ///
  /// Returns the completer future.
  Future<List<T>> onComplete() {
    return _completer.future;
  }

  /// Constructor
  AsyncProcesses(this._futures) {
    if (_futures.length == 0) {
      // Complete now if there are no futures.
      _completer.complete([]);
    } else {
      // Set completion callbacks.
      for (final future in _futures) {
        future.then((T returnValue) {
          _completed.add(returnValue);
          _futures.remove(future);

          // Complete _completer if there are no future responses left.
          if (_futures.isEmpty) {
            _completer.complete(_completed);
          }
        });
      }
    }
  }
}

/// Asynchronous Omnibus responses helper
///
/// This class helps to await multiple asynchronous Omnibus responses.
class AsyncResponses extends AsyncProcesses<OmnibusResponse> {
  /// Loop through all responses with the given omnibusId. This method returns
  /// a Future that completes once all responses are processed.
  Future forEach(String omnibusId, void f(Map<String, dynamic> data)) async {
    // Loop through all evaluated responses.
    for (final response in _completed) {
      if (response.omnibusId == omnibusId) {
        f(response.data);
      }
    }

    // Loop through all future responses and set callbacks.
    for (final futureResponse in _futures) {
      futureResponse.then((OmnibusResponse response) {
        // Note that the standard response completion logic is already handled
        // in the constructor.
        if (response.omnibusId == omnibusId) {
          f(response.data);
        }
      });
    }

    return _completer.future;
  }

  AsyncResponses(List<Future<OmnibusResponse>> responses) : super(responses);
}
