// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibus;

/// Handler pool base template
///
/// The handler pool manages handlers and condition mapping. Handlers have to be
/// added and triggered through this class.
class _HandlerPool<T extends Function> {
  /// Handler functions
  List<T> _handlers = new List<T>();

  /// Condition-less handlers
  Map<String, Set<int>> _conditionLess = new Map<String, Set<int>>();

  /// Pre-processed condition map for handlers
  Map<String, _HandlerIndex> _conditionMap = new Map<String, _HandlerIndex>();

  /// Insert new handler
  void insert(String omnibusId, Map<String, dynamic> conditions, T handler) {
    // Check if handler exists.
    if (handler != null) {
      // Insert new handler.
      int index = _handlers.length;
      _handlers.add(handler);

      // Check if the condition map exists and contains items.
      if (conditions != null && conditions.isNotEmpty) {
        // Insert index in global condition map.
        _conditionMap.putIfAbsent(omnibusId, () => new _HandlerIndex());
        _conditionMap[omnibusId].insertIndex(index, conditions);
      } else {
        // Insert handler in condition-less handler list.
        _conditionLess.putIfAbsent(omnibusId, () => new Set<int>());
        _conditionLess[omnibusId].add(index);
      }
    } else {
      throw new ArgumentError.notNull('handler');
    }
  }

  /// Retieve handler indices that match all given conditions.
  Set<int> getIndices(String omnibusId, Map<String, dynamic> messageData) {
    // Include conditionLess if available.
    Set<int> indices = _conditionLess.containsKey(omnibusId)
        ? _conditionLess[omnibusId]
        : new Set<int>();

    if (messageData != null && messageData.isNotEmpty) {
      // Check if a condition map exists.
      if (_conditionMap.containsKey(omnibusId)) {
        return indices.union(_conditionMap[omnibusId].findIndices(messageData));
      } else {
        return indices;
      }
    } else {
      // Note: If the message data is empty this message could still be used as
      // a very simple trigger.
      return indices;
    }
  }
}

/// Function for handling event messages
typedef Future EventHandler(OmnibusEvent event);

/// _HandlerPool for EventHandlers
class _EventHandlerPool extends _HandlerPool<EventHandler> {
  /// Global event handlers
  List<EventHandler> globalHandlers = new List<EventHandler>();

  /// Publish
  ///
  /// This method runs in the future and all handlers are executed
  /// asynchronously.
  Future<AsyncProcesses<Null>> publish(OmnibusEvent event) async {
    var indices = getIndices(event.omnibusId, event.data);
    var futures = new List<Future<Null>>();

    // Run global event handlers.
    for (final handler in globalHandlers) {
      futures.add(handler(event));
    }

    // Run event handlers.
    for (final i in indices) {
      futures.add(_handlers[i](event));
    }

    return new AsyncProcesses<Null>(futures);
  }
}

/// Function for handling request messages
typedef Future<OmnibusResponse> RequestHandler(OmnibusRequest request);

/// Function for handling request messages by a global handler
typedef Future<List<Future<OmnibusResponse>>> GlobalRequestHandler(
    OmnibusRequest request);

/// _HandlerPool for RequestHandlers
class _RequestHandlerPool extends _HandlerPool<RequestHandler> {
  /// Global request handlers
  List<GlobalRequestHandler> globalHandlers = new List<GlobalRequestHandler>();

  /// Publish
  ///
  /// This method runs in the future and all handlers are executed
  /// asynchronously.
  Future<AsyncResponses> publish(OmnibusRequest request) async {
    var indices = getIndices(request.omnibusId, request.data);
    var replies = new List<Future<OmnibusResponse>>();

    // Collect replies from global request handlers.
    for (final handler in globalHandlers) {
      replies.addAll(await handler(request));
    }

    // Collect replies.
    for (final i in indices) {
      replies.add(_handlers[i](request));
    }

    return new AsyncResponses(replies);
  }
}
