// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibus.json_data;

// Convert value to JSON compatible type.
dynamic getJsonValue(dynamic value) {
  // Check if this is already a JSON primitive.
  if (isJsonPrimitive(value)) {
    return value;
  } else {
    if (!isJsonType(value)) {
      // Try again with value.toJson()
      // This is deliberately not protected by a try-catch block so that
      // exceptions will be received by the top-level code.
      value = value.toJson();
      // Throw error if this is still not JSON compatible.
      if (!isJsonType(value)) {
        throw new ArgumentError.value(value, 'value',
            'toJson method does not return a JSON compatible type');
      }
    }

    // Now value should be JSON compatible.
    if (isJsonPrimitive(value)) {
      return value;
    } else if (value is List) {
      // Validate as list if this is a list.
      return getJsonList(value);
    } else {
      // This must be a Map<String, dynamic>), so validate as map.
      return getJsonMap(value);
    }
  }
}

/// Convert data to a list that only contains JSON data types.
List<dynamic> getJsonList(final List<dynamic> data) {
  // Create new List because the input data might be a List mixin that is
  // immutable. This is for example the case with YamlList.
  var newList = new List<dynamic>();

  // Iterate through list.
  for (var i = 0; i < data.length; i++) {
    // Store validated value.
    newList.add(getJsonValue(data[i]));
  }

  return newList;
}

/// Convert data to a map that only contains JSON data types.
Map<String, dynamic> getJsonMap(final Map<String, dynamic> data) {
  // Create new Map because the input data might be a Map mixin that is
  // immutable. This is for example the case with YamlMap.
  var newMap = new Map<String, dynamic>();

  // Iterate through map.
  data.forEach((String key, dynamic value) {
    // Store validated value.
    newMap[key] = getJsonValue(value);
  });

  return newMap;
}

/// Check if the given object is a JSON primitive (number, boolean, null,
/// string). This code is derived from dart:convert's json.dart
bool isJsonPrimitive(final dynamic object) {
  return (object is num ||
      identical(object, true) ||
      identical(object, false) ||
      object == null ||
      object is String);
}

/// Check if the given object is a JSON compatible type (number, boolean, null,
/// string). This code is derived from dart:convert's json.dart
bool isJsonType(final dynamic object) {
  return (object is num ||
      identical(object, true) ||
      identical(object, false) ||
      object == null ||
      object is String ||
      object is List ||
      object is Map<String, dynamic>);
}
